"""
# Web UI

Ragna comes with a web UI where you can try out all of the features!

!!! tip "Under development"

    This documentation page is being actively worked on.
    It will be updated with more details soon. :)
"""

# %%
# You can launch the web application from the CLI:
#
# ```bash
# ragna ui
# ```
#
# It will open in a browser window automatically and you should see the login screen:
#
# ![ragna UI login screen with a username and password form. Both the username and
# password are filled with the string "ragna" and the password is visible and not
# censored](../../assets/images/web-ui-login.png)
#
# By default [ragna.deploy.RagnaDemoAuthentication][] is used. Thus, you can use any
# username and a matching password to authenticate. For example, you can leave both fields
# blank or use `ragna` / `ragna` as in the image above.
